package SecureDistributedKeyGenerationForDiscreteLogBasedCryptosystem;

import Communication.Network;
import JointFeldmanProtocol.DistributedKeyGenerationUserImpl;
import ShamirSecretSharing.Polynomial;
import ShamirSecretSharing.SecretSharing;
import com.google.protobuf.Message;
import meerkat.protobuf.DKGMessages;

import java.math.BigInteger;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;

/**
 * Created by Tzlil on 2/22/2016.
 */
public class SecureDistributedKeyGenerationUserImpl extends DistributedKeyGenerationUserImpl {

    private final SecureDistributedKeyGeneration sdkg;
    private final Polynomial.Point[] sharesT;
    private final BigInteger[][] verificationValuesTable;
    private final Hashtable<Integer,Set<Polynomial.Point>> ysRestoreShares;

    public SecureDistributedKeyGenerationUserImpl(SecureDistributedKeyGeneration sdkg, Network network) {
        super(sdkg, network,new SecureDistributedKeyGenerationMailHandler(null));
        this.sdkg = sdkg;
        this.sharesT = new Polynomial.Point[n];
        this.verificationValuesTable = new BigInteger[n][t + 1];
        this.ysRestoreShares = new Hashtable<Integer, Set<Polynomial.Point>>();
        this.messageHandler = new MessageHandler();
        this.user.setMessageHandler(this.messageHandler);
    }

    /**
     * stage1 according to the protocol
     *  1. Pi broadcasts Cik=Aik*Bik for k = 0,...,t.
     *  2. Pi computes the shares Sij,Sij' for j = 1,...,n and sends Sij,Sij' secretly to Pj.
     */
    @Override
    protected void stage1() {
        sdkg.broadcastVerificationValues(user);
        sdkg.sendSecrets(user);
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares,sharesT he received
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     *  Pj broadcasts done message at the end of this stage
     */
    @Override
    protected void stage2(){
        sdkg.setShares(shares);
        sdkg.setSharesT(sharesT);
        sdkg.broadcastComplains(user,verificationValuesTable);
        //broadcast done message after all complaints
        DKGMessages.DoneMessage doneMessage = DKGMessages.DoneMessage.newBuilder().build();
        user.broadcast(DKGMessages.Mail.Type.DONE,doneMessage);
    }

    private void ys(){
        sdkg.broadcastCommitments(user);
        //wait for receive all commitments from all i in QUAL
        for (int i:QUAL) {
            for(int k = 0; k <= t; k++) {
                while (commitmentsTable[i - 1][k] == null) {
                    try {
                        Thread.sleep(300);
                    } catch (InterruptedException e) {
                        // do nothing
                    }
                }
            }
        }
        sdkg.broadcastComplaints(user,commitmentsTable,true);
        //broadcast done message after all complaints
        DKGMessages.DoneMessage doneMessage = DKGMessages.DoneMessage.newBuilder().build();
        user.broadcast(DKGMessages.Mail.Type.DONE,doneMessage);

        for (int i:QUAL) {
            while (doneFlags[i - 1]) {
                try {
                    Thread.sleep(300);
                } catch (InterruptedException e) {
                    // do nothing
                }
            }
        }
        BigInteger secret;
        for (Integer i: ysRestoreShares.keySet()) {
            try {
                 secret = SecretSharing.restoreSecret((Polynomial.Point[])ysRestoreShares.get(i).toArray());
                 //ToDo use restored secret...
            } catch (Exception e) {
                //
            }
        }
    }

    @Override
    protected void stage4() {
        isStage4 = true;
        ((SecureDistributedKeyGenerationMailHandler)user.getMailHandler()).setStage4(true);
        ys();
        super.stage4();
    }
    boolean isStage4 = false;
    private class MessageHandler extends DistributedKeyGenerationUserImpl.MessageHandler{

        final int NumberOfCommitmentsInStage1 = n * (t + 1);

        @Override
        protected boolean isValidCommitmentMessage(int sender, boolean isBroadcast, DKGMessages.CommitmentMessage commitmentMessage) {
            if(commitmentsCounter < NumberOfCommitmentsInStage1) {
                int i = sender - 1;
                int k = commitmentMessage.getK();
                return isBroadcast && verificationValuesTable[i][k] == null;
            }else {
                return super.isValidCommitmentMessage(sender, isBroadcast, commitmentMessage);
            }
        }

        @Override
        public void handelCommitmentMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.CommitmentMessage commitmentMessage = ( DKGMessages.CommitmentMessage)message;
            if(commitmentsCounter < NumberOfCommitmentsInStage1) {
                if(isValidCommitmentMessage(sender,isBroadcast,commitmentMessage)) {
                    int i = sender - 1;
                    int k = commitmentMessage.getK();
                    verificationValuesTable[i][k] = extractCommitment(commitmentMessage);
                    commitmentsCounter++;
                }
            }
            else{
                super.handelCommitmentMessage(sender,isBroadcast,commitmentMessage);
            }
        }

        protected boolean isValidSecretMessage(int sender, boolean isBroadcast, DKGMessages.DoubleSecretMessage doubleSecretMessage) {
            DKGMessages.SecretMessage secretMessage = DKGMessages.SecretMessage.newBuilder()
                    .setI(doubleSecretMessage.getI())
                    .setJ(doubleSecretMessage.getJ())
                    .setSecret(doubleSecretMessage.getSecret())
                    .build();
            return super.isValidSecretMessage(sender,isBroadcast,secretMessage);
        }

        @Override
        public void handelSecretMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.DoubleSecretMessage doubleSecretMessage = (DKGMessages.DoubleSecretMessage)message;
            if (isValidSecretMessage(sender,isBroadcast,doubleSecretMessage)) {
                int i = doubleSecretMessage.getI();

                Polynomial.Point secret = extractSecret(i, doubleSecretMessage.getSecret());
                Polynomial.Point secretT = extractSecret(i, doubleSecretMessage.getSecretT());
                shares[i - 1] = secret;
                sharesT[i - 1] = secretT;
                secretsCounter++;
            }
        }
        protected boolean isValidAnswerMessage(int sender, boolean isBroadcast, DKGMessages.DoubleSecretMessage doubleSecretMessage) {
            if(!isStage4) {
                DKGMessages.SecretMessage secretMessage = DKGMessages.SecretMessage.newBuilder()
                        .setI(doubleSecretMessage.getI())
                        .setJ(doubleSecretMessage.getJ())
                        .setSecret(doubleSecretMessage.getSecret())
                        .build();
                return super.isValidAnswerMessage(sender, isBroadcast, secretMessage);
            }else{
                int i = doubleSecretMessage.getI();
                int j = doubleSecretMessage.getJ();
                return isBroadcast && j == sender && QUAL.contains(i) && QUAL.contains(j)&& ysRestoreShares.containsKey(i);
            }
        }

        @Override
        public void handelAnswerMessage(int sender, boolean isBroadcast, Message message) {
            DKGMessages.DoubleSecretMessage doubleSecretMessage = (DKGMessages.DoubleSecretMessage)message;
            if(isValidAnswerMessage(sender,isBroadcast,doubleSecretMessage)) {
                int i = doubleSecretMessage.getI();
                int j = doubleSecretMessage.getJ();
                Polynomial.Point secret = extractSecret(i, doubleSecretMessage.getSecret());
                Polynomial.Point secretT = extractSecret(i, doubleSecretMessage.getSecretT());
                if (!isStage4) {
                    if (sdkg.isValidSecret(secret, secretT, verificationValuesTable[j - 1], i)) {
                        complaintsTable[i - 1][j - 1] = ComplainState.NonDisqualified;
                    } else {
                        complaintsTable[i - 1][j - 1] = ComplainState.Disqualified;
                    }
                } else {
                    if (ysRestoreShares.get(i).add(secret) && sender != id) {
                        sdkg.broadcastAnswer(user, secret, secretT, i);
                    }
                }
            }
        }

        @Override
        protected boolean isValidDoneMessage(int sender, boolean isBroadcast) {
            if(doneCounter < n) {
                return super.isValidDoneMessage(sender, isBroadcast);
            }else{
                return isBroadcast && doneFlags[sender - 1];
            }
        }

        @Override
        public void handelDoneMessage(int sender, boolean isBroadcast, Message message) {
            if(doneCounter < n)
                super.handelDoneMessage(sender, isBroadcast, message);
            else{
                if(isValidDoneMessage(sender,isBroadcast)) {
                    doneFlags[sender - 1] = false;
                }
            }
        }

        protected boolean isValidComplaintMessage(int sender, boolean isBroadcast,
                                                  DKGMessages.DoubleSecretMessage ysComplaintMessage){
            int i = ysComplaintMessage.getI();
            int j = ysComplaintMessage.getJ();
            return isBroadcast && j == sender && QUAL.contains(i) && QUAL.contains(j)&&!ysRestoreShares.containsKey(i);
        }

        @Override
        public void handelComplaintMessage(int sender, boolean isBroadcast, Message message) {
            if(!isStage4) {
                super.handelComplaintMessage(sender, isBroadcast, message);
            }else {
                DKGMessages.DoubleSecretMessage ysComplaintMessage =(DKGMessages.DoubleSecretMessage)message;
                if (isValidComplaintMessage(sender,isBroadcast,ysComplaintMessage)) {
                    int i = ysComplaintMessage.getI();
                    int j = ysComplaintMessage.getJ();
                    Polynomial.Point secret = extractSecret(i,ysComplaintMessage.getSecret());
                    Polynomial.Point secretT = extractSecret(i,ysComplaintMessage.getSecretT());
                    if (sdkg.isValidSecret(secret, secretT, verificationValuesTable[i - 1], j)
                            && !sdkg.isValidSecret(secret, commitmentsTable[i - 1], j)) {
                        ysRestoreShares.put(i, new HashSet<Polynomial.Point>());
                        ysRestoreShares.get(i).add(secret);
                        sdkg.broadcastAnswer(user, secret, secretT, i);
                    }
                }
            }
        }
    }
}
